@@ -45,10 +45,12 @@ static
 void
 past_node_init (s_past_node_t* node,
 		cs_past_node_type_t* type,
-		past_visitor_fun_t visitor)
+		past_visitor_fun_t visitor,
+		past_clone_fun_t clone)
 {
   node->type = type;
   node->visitor = visitor;
+  node->clone = clone;
   node->parent = NULL;
   node->next = NULL;
   node->metainfo = NULL;
@@ -83,7 +85,7 @@ past_root_free (s_past_node_t* node)
 {
   assert (past_node_is_a (node, past_root));
   s_past_root_t* pr = (s_past_root_t*) node;
-  symbol_table_free (pr->symboltable);
+  //symbol_table_free (pr->symboltable);
   past_node_freeattr (node);
   XFREE(pr);
 }
@@ -100,11 +102,19 @@ past_root_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(root);
 
+static s_past_node_t*
+past_root_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_root));
+  PAST_DECLARE_TYPED(root, r, node);
+  return past_node_root_create (r->symboltable, past_clone (r->body));
+}
+
 s_past_root_t* past_root_create (s_symbol_table_t* symboltable,
 				 s_past_node_t* body)
 {
   s_past_root_t* n = XMALLOC(s_past_root_t, 1);
-  past_node_init (&(n->node), past_root, past_root_visitor);
+  past_node_init (&(n->node), past_root, past_root_visitor, past_root_clone);
   n->symboltable = symboltable;
   n->body = body;
   set_parent_sibling_nodes (n->body, (s_past_node_t*) n);
@@ -143,10 +153,27 @@ past_variable_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(variable);
 
+static s_past_node_t*
+past_variable_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_variable));
+  PAST_DECLARE_TYPED(variable, v, node);
+  s_symbol_t* newsym = NULL;
+  if (v->symbol->is_char_data)
+    newsym = symbol_add_from_char (NULL, v->symbol->data);
+  else
+    newsym = symbol_add_from_data (NULL, v->symbol->data);
+  s_past_node_t* n =  past_node_variable_create (newsym);
+  PAST_DECLARE_TYPED(variable, pv, n);
+  n->usr = v->usr;
+  return n;
+}
+
 s_past_variable_t* past_variable_create (s_symbol_t* symbol)
 {
   s_past_variable_t* n = XMALLOC(s_past_variable_t, 1);
-  past_node_init (&(n->node), past_variable, past_variable_visitor);
+  past_node_init (&(n->node), past_variable, past_variable_visitor,
+		  past_variable_clone);
   n->symbol = symbol;
   n->usr = NULL;
 
@@ -183,10 +210,18 @@ past_value_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(value);
 
+static s_past_node_t*
+past_value_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_value));
+  PAST_DECLARE_TYPED(value, v, node);
+  return past_node_value_create (v->type, v->value);
+}
+
 s_past_value_t* past_value_create (int type, u_past_value_data_t val)
 {
   s_past_value_t* n = XMALLOC(s_past_value_t, 1);
-  past_node_init (&(n->node), past_value, past_value_visitor);
+  past_node_init (&(n->node), past_value, past_value_visitor, past_value_clone);
   n->type = type;
   n->value = val;
 
@@ -245,6 +280,25 @@ past_for_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(for);
 
+static s_past_node_t*
+past_for_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_for));
+  PAST_DECLARE_TYPED(for, f, node);
+  s_past_node_t* n = past_node_for_create (past_clone (f->init),
+					   past_clone (f->test),
+					   (s_past_variable_t*)
+					   f->iterator->node.clone
+					   (&(f->iterator->node)),
+					   past_clone(f->increment),
+					   past_clone (f->body),
+					   f->usr);
+  PAST_DECLARE_TYPED(for, cf, n);
+  cf->type = f->type;
+
+  return n;
+}
+
 s_past_for_t* past_for_create (s_past_node_t* init,
 			       s_past_node_t* test,
 			       s_past_variable_t* iterator,
@@ -253,7 +307,7 @@ s_past_for_t* past_for_create (s_past_node_t* init,
 			       void* usr)
 {
   s_past_for_t* n = XMALLOC(s_past_for_t, 1);
-  past_node_init (&(n->node), past_for, past_for_visitor);
+  past_node_init (&(n->node), past_for, past_for_visitor, past_for_clone);
   n->init = init;
   n->test = test;
   n->iterator = iterator;
@@ -317,6 +371,40 @@ past_parfor_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_2(parfor,for);
 
+static s_past_node_t*
+past_parfor_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_parfor));
+  PAST_DECLARE_TYPED(parfor, f, node);
+  s_past_node_t* n = past_node_parfor_create (past_clone (f->init),
+					      past_clone (f->test),
+					      (s_past_variable_t*)
+					      f->iterator->node.clone
+					      (&(f->iterator->node)),
+					      past_clone(f->increment),
+					      past_clone (f->body),
+					      f->usr);
+  PAST_DECLARE_TYPED(parfor, cf, n);
+  cf->type = f->type;
+  if (f->private_vars)
+    {
+      int i;
+      for (i = 0; f->private_vars[i]; ++i)
+	;
+      s_past_variable_t** newvars = XMALLOC(s_past_variable_t*, i + 1);
+      for (i = 0; f->private_vars[i]; ++i)
+	{
+	  s_past_node_t* newvar =
+	    past_clone ((s_past_node_t*) f->private_vars[i]);
+	  newvars[i] = (s_past_variable_t*) newvar;
+	}
+      newvars[i] = NULL;
+      cf->private_vars = newvars;
+    }
+
+  return n;
+}
+
 s_past_parfor_t* past_parfor_create (s_past_node_t* init,
 				     s_past_node_t* test,
 				     s_past_variable_t* iterator,
@@ -325,7 +413,8 @@ s_past_parfor_t* past_parfor_create (s_past_node_t* init,
 				     void* usr)
 {
   s_past_parfor_t* n = XMALLOC(s_past_parfor_t, 1);
-  past_node_init (&(n->node), past_parfor, past_parfor_visitor);
+  past_node_init (&(n->node), past_parfor, past_parfor_visitor,
+		  past_parfor_clone);
   n->init = init;
   n->test = test;
   n->iterator = iterator;
@@ -339,7 +428,7 @@ s_past_parfor_t* past_parfor_create (s_past_node_t* init,
   set_parent_sibling_nodes (n->increment, (s_past_node_t*) n);
   set_parent_sibling_nodes (n->body, (s_past_node_t*) n);
 
-  return (s_past_parfor_t*)n;
+  return n;
 }
 
 s_past_node_t* past_node_parfor_create (s_past_node_t* init,
@@ -412,15 +501,26 @@ declare_bin_free_ops(gt);
 declare_bin_free_ops(lt);
 declare_bin_free_ops(arrayref);
 
+
+static s_past_node_t*
+past_binary_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_binary));
+  PAST_DECLARE_TYPED(binary, b, node);
+  return past_node_binary_create (node->type, past_clone (b->lhs),
+				  past_clone (b->rhs));
+}
+
 s_past_binary_t* past_binary_create (cs_past_node_type_t* type,
 				     s_past_node_t* lhs,
 				     s_past_node_t* rhs)
 {
   s_past_binary_t* n = XMALLOC(s_past_binary_t, 1);
   if (type)
-    past_node_init (&(n->node), type, past_binary_visitor);
+    past_node_init (&(n->node), type, past_binary_visitor, past_binary_clone);
   else
-    past_node_init (&(n->node), past_binary, past_binary_visitor);
+    past_node_init (&(n->node), past_binary, past_binary_visitor,
+		    past_binary_clone);
   n->lhs = lhs;
   n->rhs = rhs;
   set_parent_sibling_nodes (n->lhs, (s_past_node_t*) n);
@@ -484,14 +584,23 @@ declare_un_free_ops(inc_after);
 declare_un_free_ops(dec_before);
 declare_un_free_ops(dec_after);
 
+static s_past_node_t*
+past_unary_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_unary));
+  PAST_DECLARE_TYPED(unary, u, node);
+  return past_node_unary_create (node->type, past_clone (u->expr));
+}
+
 s_past_unary_t* past_unary_create (cs_past_node_type_t* type,
 				   s_past_node_t* expr)
 {
   s_past_unary_t* n = XMALLOC(s_past_unary_t, 1);
   if (type)
-    past_node_init (&(n->node), type, past_unary_visitor);
+    past_node_init (&(n->node), type, past_unary_visitor, past_unary_clone);
   else
-    past_node_init (&(n->node), past_unary, past_unary_visitor);
+    past_node_init (&(n->node), past_unary, past_unary_visitor,
+		    past_unary_clone);
   n->expr = expr;
   set_parent_sibling_nodes (n->expr, (s_past_node_t*) n);
 
@@ -532,10 +641,18 @@ past_block_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(block);
 
+static s_past_node_t*
+past_block_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_block));
+  PAST_DECLARE_TYPED(block, b, node);
+  return past_node_block_create (past_clone (b->body));
+}
+
 s_past_block_t* past_block_create (s_past_node_t* body)
 {
   s_past_block_t* n = XMALLOC(s_past_block_t, 1);
-  past_node_init (&(n->node), past_block, past_block_visitor);
+  past_node_init (&(n->node), past_block, past_block_visitor, past_block_clone);
   n->body = body;
   set_parent_sibling_nodes (n->body, (s_past_node_t*) n);
 
@@ -579,12 +696,22 @@ past_if_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(if);
 
+static s_past_node_t*
+past_if_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_if));
+  PAST_DECLARE_TYPED(if, b, node);
+  return past_node_if_create (past_clone (b->condition),
+			      past_clone (b->then_clause),
+			      past_clone (b->else_clause));
+}
+
 s_past_if_t* past_if_create (s_past_node_t* cond,
 			     s_past_node_t* then_clause,
 			     s_past_node_t* else_clause)
 {
   s_past_if_t* n = XMALLOC(s_past_if_t, 1);
-  past_node_init (&(n->node), past_if, past_if_visitor);
+  past_node_init (&(n->node), past_if, past_if_visitor, past_if_clone);
   n->condition = cond;
   n->then_clause = then_clause;
   n->else_clause = else_clause;
@@ -632,11 +759,21 @@ past_affineguard_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(affineguard);
 
+static s_past_node_t*
+past_affineguard_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_affineguard));
+  PAST_DECLARE_TYPED(affineguard, b, node);
+  return past_node_affineguard_create (past_clone (b->condition),
+				       past_clone (b->then_clause));
+}
+
 s_past_affineguard_t* past_affineguard_create (s_past_node_t* cond,
 					       s_past_node_t* then_clause)
 {
   s_past_affineguard_t* n = XMALLOC(s_past_affineguard_t, 1);
-  past_node_init (&(n->node), past_affineguard, past_affineguard_visitor);
+  past_node_init (&(n->node), past_affineguard, past_affineguard_visitor,
+		  past_affineguard_clone);
   n->condition = cond;
   n->then_clause = then_clause;
   set_parent_sibling_nodes (n->condition, (s_past_node_t*) n);
@@ -682,6 +819,19 @@ past_cloogstmt_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(cloogstmt);
 
+static s_past_node_t*
+past_cloogstmt_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_cloogstmt));
+  PAST_DECLARE_TYPED(cloogstmt, b, node);
+  return past_node_cloogstmt_create (b->cloogdomain, b->cloogstatement,
+				     (s_past_variable_t*)
+				     b->stmt_name->node.clone
+				     (&(b->stmt_name->node)),
+				     b->stmt_number,
+				     past_clone (b->substitutions));
+}
+
 s_past_cloogstmt_t* past_cloogstmt_create (void* cloogdomain,
 					   void* cloogstatement,
 					   s_past_variable_t* stmt_name,
@@ -689,7 +839,8 @@ s_past_cloogstmt_t* past_cloogstmt_create (void* cloogdomain,
 					   s_past_node_t* substitutions)
 {
   s_past_cloogstmt_t* n = XMALLOC(s_past_cloogstmt_t, 1);
-  past_node_init (&(n->node), past_cloogstmt, past_cloogstmt_visitor);
+  past_node_init (&(n->node), past_cloogstmt, past_cloogstmt_visitor,
+		  past_cloogstmt_clone);
   n->cloogdomain = cloogdomain;
   n->cloogstatement = cloogstatement;
   n->stmt_name = stmt_name;
@@ -740,10 +891,19 @@ past_statement_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(statement);
 
+static s_past_node_t*
+past_statement_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_statement));
+  PAST_DECLARE_TYPED(statement, b, node);
+  return past_node_statement_create (past_clone (b->body));
+}
+
 s_past_statement_t* past_statement_create (s_past_node_t* body)
 {
   s_past_statement_t* n = XMALLOC(s_past_statement_t, 1);
-  past_node_init (&(n->node), past_statement, past_statement_visitor);
+  past_node_init (&(n->node), past_statement, past_statement_visitor,
+		  past_statement_clone);
   n->body = body;
   set_parent_sibling_nodes (n->body, (s_past_node_t*) n);
 
@@ -784,11 +944,22 @@ past_funcall_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(funcall);
 
+static s_past_node_t*
+past_funcall_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_funcall));
+  PAST_DECLARE_TYPED(funcall, b, node);
+  return past_node_funcall_create ((s_past_variable_t*)
+				   b->name->node.clone (&(b->name->node)),
+				   past_clone (b->args_list));
+}
+
 s_past_funcall_t* past_funcall_create (s_past_variable_t* name,
 				       s_past_node_t* args_list)
 {
   s_past_funcall_t* n = XMALLOC(s_past_funcall_t, 1);
-  past_node_init (&(n->node), past_funcall, past_funcall_visitor);
+  past_node_init (&(n->node), past_funcall, past_funcall_visitor,
+		  past_funcall_clone);
   n->name = name;
   n->args_list = args_list;
 
@@ -829,11 +1000,23 @@ past_vardecl_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(vardecl);
 
+static s_past_node_t*
+past_vardecl_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_vardecl));
+  PAST_DECLARE_TYPED(vardecl, b, node);
+  return past_node_vardecl_create ((s_past_variable_t*)
+				   b->name->node.clone (&(b->name->node)),
+				   (s_past_variable_t*)
+				   b->type->node.clone (&(b->type->node)));
+}
+
 s_past_vardecl_t* past_vardecl_create (s_past_variable_t* name,
 				       s_past_variable_t* type)
 {
   s_past_vardecl_t* n = XMALLOC(s_past_vardecl_t, 1);
-  past_node_init (&(n->node), past_vardecl, past_vardecl_visitor);
+  past_node_init (&(n->node), past_vardecl, past_vardecl_visitor,
+		  past_vardecl_clone);
   n->name = name;
   n->type = type;
 
@@ -879,12 +1062,23 @@ past_ternary_cond_visitor (s_past_node_t* node,
 }
 PAST_DECLARE_NODE_IN_HIERARCHY_UNIT_1(ternary_cond);
 
+static s_past_node_t*
+past_ternary_cond_clone (s_past_node_t* node)
+{
+  assert (past_node_is_a (node, past_ternary_cond));
+  PAST_DECLARE_TYPED(ternary_cond, b, node);
+  return past_node_ternary_cond_create (past_clone (b->cond),
+					past_clone (b->true_clause),
+					past_clone (b->false_clause));
+}
+
 s_past_ternary_cond_t* past_ternary_cond_create (s_past_node_t* cond,
 						 s_past_node_t* true_clause,
 						 s_past_node_t* false_clause)
 {
   s_past_ternary_cond_t* n = XMALLOC(s_past_ternary_cond_t, 1);
-  past_node_init (&(n->node), past_ternary_cond, past_ternary_cond_visitor);
+  past_node_init (&(n->node), past_ternary_cond, past_ternary_cond_visitor,
+		  past_ternary_cond_clone);
   n->cond = cond;
   n->true_clause = true_clause;
   n->false_clause = false_clause;
@@ -902,12 +1096,35 @@ s_past_node_t* past_node_ternary_cond_create (s_past_node_t* cond,
 
 
 /**
- *
  * cloning a node/tree (deep copy).
  *
- *
  */
 s_past_node_t* past_clone (s_past_node_t* s)
+{
+  if (! s)
+    return NULL;
+
+  s_past_node_t* ret = NULL;
+  s_past_node_t* val;
+  s_past_node_t** stmt = &val;
+
+  for ( ; s; s = s->next)
+    {
+      *stmt = s->clone (s);
+      if (ret == NULL)
+	ret = *stmt;
+      stmt = &((*stmt)->next);
+    }
+  *stmt = NULL;
+
+  return ret;
+}
+
+/**
+ * Old code. Unused now.
+ */
+static
+s_past_node_t* past_clone_old (s_past_node_t* s)
 {
   if (! s)
     return NULL;
