@@ -42,6 +42,8 @@
 #include <string.h>
 #include <candl/candl.h>
 
+#include <candl/piplib-wrapper.h>
+
 #include <assert.h>
 
 #ifdef CANDL_SUPPORTS_ISL
@@ -52,6 +54,8 @@
 # include <isl/set.h>
 #endif
 
+
+
 /******************************************************************************
  *                          Structure display function                        *
  ******************************************************************************/
@@ -110,6 +114,7 @@ void candl_dependence_print_structure(FILE* file,
 	case CANDL_WAR   : fprintf(file, "WAR (anti)\n");   break;
 	case CANDL_WAW   : fprintf(file, "WAW (output)\n"); break;
 	case CANDL_RAR   : fprintf(file, "RAR (input)\n");  break;
+	case CANDL_RAW_SCALPRIV   : fprintf(file, "RAW_SCALPRIV (scalar priv)\n");  break;
 	default : fprintf(file, "unknown\n"); break;
 	}
 
@@ -189,6 +194,7 @@ void candl_dependence_pprint(FILE * file, candl_dependence_p dependence)
 	case CANDL_WAR   : fprintf(file, "WAR")  ; break;
 	case CANDL_WAW   : fprintf(file, "WAW")  ; break;
 	case CANDL_RAR   : fprintf(file, "RAR")  ; break;
+	case CANDL_RAW_SCALPRIV   : fprintf(file, "RAW_SCALPRIV (scalar-priv)")  ; break;
 	default : fprintf(file, "unknown"); break;
 	}
       fprintf(file, " depth %d, ref %d->%d \"];\n", dependence->depth,
@@ -276,6 +282,11 @@ candl_program_deps_to_string(CandlDependence* dependence)
 	      refs = CANDL_get_si(tmp->source->read->p[tmp->ref_source][0]);
 	      reft = CANDL_get_si(tmp->target->read->p[tmp->ref_target][0]);
 	      break;
+	    case CANDL_RAW_SCALPRIV:
+	      type = "RAW_SCALPRIV #(scalar priv)";
+	      refs = CANDL_get_si(tmp->source->written->p[tmp->ref_source][0]);
+	      reft = CANDL_get_si(tmp->target->read->p[tmp->ref_target][0]);
+	      break;
 	    default:
 	      exit(1);
 	      break;
@@ -425,6 +436,8 @@ CandlDependence* candl_dependence_read_one_dep(char* str, char** next,
     dep->type = CANDL_WAR;
   else if (! strcmp(buffer, "WAW"))
     dep->type = CANDL_WAW;
+  else if (! strcmp(buffer, "RAW_SCALPRIV"))
+    dep->type = CANDL_RAW_SCALPRIV;
   for (; *str != '\n'; ++str);
   ++str;
 
@@ -482,6 +495,7 @@ CandlDependence* candl_dependence_read_one_dep(char* str, char** next,
   switch (dep->type)
     {
     case CANDL_RAW:
+    case CANDL_RAW_SCALPRIV:
       msource = dep->source->written;
       mtarget = dep->target->read;
       break;
@@ -910,16 +924,17 @@ int candl_dependence_gcd_test(CandlStatement* source,
   int gcd;
   int id;
   int value;
-  int null_iter, null_param, null_cst, pos_iter, neg_iter, strict_pred;
+  int null_iter, null_param, null_cst, pos_iter, neg_iter;
 
   /* Check that the precedence constraint, if any, is not strict in a
      self-dependence. */
-  if (source == target &&
-      CANDL_get_si(system->p[system->NbRows - 1][0]) == 1 &&
-      CANDL_get_si(system->p[system->NbRows - 1][system->NbColumns - 1]) == -1)
-    strict_pred = 1;
-  else
-    strict_pred = 0;
+  /* int strict_pred; */
+  /* if (source == target && */
+  /*     CANDL_get_si(system->p[system->NbRows - 1][0]) == 1 && */
+  /*     CANDL_get_si(system->p[system->NbRows - 1][system->NbColumns - 1]) == -1) */
+  /*   strict_pred = 1; */
+  /* else */
+  /*   strict_pred = 0; */
 
   /* Inspect the array access function equalities. */
   for (id = source->domain->NbRows + target->domain->NbRows;
@@ -1170,22 +1185,8 @@ candl_dependence_p candl_dependence_system(CandlStatement* source,
   if (! candl_dependence_gcd_test(source, target, system, depth))
     return NULL;
 
-#ifdef CANDL_HAS_PIPLIB_HYBRID
-  if (piplib_hybrid_has_rational_point (system, context, 1))
+  if (pip_has_rational_point (system, context, 1))
     {
-#else
-  PipOptions * options;
-  PipQuast * solution;
-  options = pip_options_init();
-  options->Simplify = 1;
-  options->Urs_parms = -1;
-  options->Urs_unknowns = -1;
-  solution = pip_solve(system,context, -1, options);
-
-  if ((solution != NULL) &&
-      ((solution->list != NULL) || (solution->condition != NULL)))
-    {
-#endif
       dependence = candl_dependence_malloc();
 
       /* We set the various fields with corresponding values. */
@@ -1197,14 +1198,8 @@ candl_dependence_p candl_dependence_system(CandlStatement* source,
       dependence->ref_target = ref_t;
       dependence->domain     = system;
     }
-  else {
+  else
     candl_matrix_free(system);
-  }
-
-#ifndef CANDL_HAS_PIPLIB_HYBRID
-  pip_options_free(options);
-  pip_quast_free(solution);
-#endif
 
   return dependence;
 }
@@ -1366,9 +1361,10 @@ candl_dependence_p candl_dependence(candl_program_p program,
     }
 
   /* If scalar analysis is called, remove some useless dependences. */
-  if (options->scalar_privatization || options->scalar_expansion ||
-      options->scalar_renaming)
-    candl_dependence_prune_scalar_waw (program, options, &dependence);
+  /* LNP: This is subsubmed by the updated prune-with-privatization function. */
+  /* if (options->scalar_privatization || options->scalar_expansion || */
+  /*     options->scalar_renaming) */
+  /*   candl_dependence_prune_scalar_waw (program, options, &dependence); */
 
   /* Final treatment for scalar analysis. */
   int check = 0;
@@ -1601,8 +1597,8 @@ candl_dependence_compute_lb (CandlMatrix* m, Entier* lb, int col)
       ((solution->list != NULL) || (solution->condition != NULL)))
     {
       l = solution->list;
-      for (; l && col > 0; l = l->next, --col)
-	;
+      while (col-- > 1)
+	l = l->next;
       CANDL_assign(*lb, l->vector->the_vector[0]);
     }
   pip_options_free (options);
@@ -1627,7 +1623,6 @@ candl_dependence_check_domain_is_included(CandlStatement* s1,
   Entier lb; CANDL_init(lb);
   max = s1->depth < max ? s1->depth : max;
   max = s2->depth < max ? s2->depth : max;
-
   CandlMatrix* m = candl_matrix_malloc(s2->domain->NbRows + s2->depth - max +1,
 				       s2->domain->NbColumns);
   int i, j;
@@ -1755,6 +1750,7 @@ candl_dependence_get_array_refs_in_dep (CandlDependence* tmp, int* refs,
       *reft = CANDL_get_si(tmp->target->read->p[tmp->ref_target][0]);
       break;
     case CANDL_RAW:
+    case CANDL_RAW_SCALPRIV:
       *refs = CANDL_get_si(tmp->source->written->p[tmp->ref_source][0]);
       *reft = CANDL_get_si(tmp->target->read->p[tmp->ref_target][0]);
       break;
@@ -2019,12 +2015,12 @@ candl_dependence_is_loop_carried (candl_program_p program,
   else
     mats = dep->source->written;
   CandlMatrix* matt;
-  if (dep->type == CANDL_RAW || dep->type == CANDL_RAR)
+  if (dep->type == CANDL_RAW || dep->type == CANDL_RAW_SCALPRIV
+      || dep->type == CANDL_RAR)
     matt = dep->target->read;
   else
     matt = dep->target->written;
 
-  int loopdep = 1;
   int src_ref_iter = 0;
   int dst_ref_iter = 0;
   for (k = 0; k == 0 ||
@@ -2032,7 +2028,6 @@ candl_dependence_is_loop_carried (candl_program_p program,
 	  CANDL_get_si(mats->p[dep->ref_source + k][0]) == 0); ++k)
     if (CANDL_get_si(mats->p[dep->ref_source + k][i + 1]) != 0)
       {
-	loopdep = 0;
 	src_ref_iter = 1;
 	break;
       }
@@ -2041,7 +2036,6 @@ candl_dependence_is_loop_carried (candl_program_p program,
 	  CANDL_get_si(matt->p[dep->ref_target + k][0]) == 0); ++k)
     if (CANDL_get_si(matt->p[dep->ref_target + k][j + 1]) != 0)
       {
-	loopdep = 0;
 	dst_ref_iter = 1;
 	break;
       }
@@ -2137,13 +2131,13 @@ candl_dependence_is_loop_carried (candl_program_p program,
   CANDL_set_si(testsyst->p[pos][testsyst->NbColumns - 1], -1);
   CANDL_set_si(testsyst->p[pos][1 + i], 1);
   CANDL_set_si(testsyst->p[pos][1 + i + src->depth], -1);
-  has_pt = piplib_hybrid_has_rational_point (testsyst, NULL, 1);
+  has_pt = pip_has_rational_point (testsyst, NULL, 1);
   if (! has_pt)
     {
       // Test for '<'.
       CANDL_set_si(testsyst->p[pos][1 + i], -1);
       CANDL_set_si(testsyst->p[pos][1 + i + src->depth], 1);
-      has_pt = piplib_hybrid_has_rational_point (testsyst, NULL, 1);
+      has_pt = pip_has_rational_point (testsyst, NULL, 1);
     }
 
   candl_matrix_free (testsyst);
@@ -2193,9 +2187,10 @@ candl_dependence_prune_with_privatization (candl_program_p program,
   CandlDependence* tmp;
   CandlDependence* pred = NULL;
   int is_priv;
-  int i;
+  int i, j;
   int loop_idx = 0;
   int refs, reft;
+  int loop_pos_priv;
 
 
   if (options->verbose)
@@ -2235,18 +2230,40 @@ candl_dependence_prune_with_privatization (candl_program_p program,
 	}
       else
 	loop_idx = tmp->source->index[i];
+      loop_pos_priv = i;
       /* Check if the dependence is loop-carried at loop i. */
       if (is_priv && candl_dependence_is_loop_carried (program, tmp, loop_idx))
 	{
-	  /* It is, the dependence can be removed. */
+	  /* If so, make the dependence loop-independent. */
+	  CandlMatrix* m = candl_matrix_malloc (tmp->domain->NbRows + 1,
+						tmp->domain->NbColumns);
+	  for (i = 0; i < tmp->domain->NbRows; ++i)
+	    for (j = 0; j < tmp->domain->NbColumns; ++j)
+	      CANDL_assign(m->p[i][j], tmp->domain->p[i][j]);
 	  candl_matrix_free (tmp->domain);
+	  tmp->domain = m;
+	  CANDL_set_si(tmp->domain->p[tmp->domain->NbRows - 1]
+		       [1 + loop_pos_priv], 1);
+	  CANDL_set_si(tmp->domain->p[tmp->domain->NbRows - 1]
+		       [1 + loop_pos_priv + tmp->source->depth], -1);
+	  /* Set the type of the dependence as special
+	     scalar-privatization one. */
+	  if (tmp->type == CANDL_RAW)
+	    tmp->type = CANDL_RAW_SCALPRIV;
 	  next = tmp->next;
-	  if (pred == NULL)
-	    *deps = next;
-	  else
-	    pred->next = next;
-	  free (tmp);
+	  if (!candl_matrix_check_point (tmp->domain, NULL))
+	    {
+	      /* It is, the dependence can be removed. */
+	      candl_matrix_free (tmp->domain);
+	      if (pred == NULL)
+		*deps = next;
+	      else
+		pred->next = next;
+	      free (tmp);
+	    }
+	  pred = tmp;
 	  tmp = next;
+
 	  continue;
 	}
       /* Go to the next victim. */
@@ -2302,7 +2319,7 @@ candl_dependence_analyze_scalars(candl_program_p program,
   CandlStatement** fullchain = NULL;
   int i, j, k, l, n;
   CandlMatrix* m;
-  int idx, max, is_priv, cpt, offset, was_priv;
+  int max, is_priv, cpt, offset, was_priv;
   CandlStatement* curlast;
   CandlStatement* last;
   int nb_priv = 0;
@@ -2322,7 +2339,6 @@ candl_dependence_analyze_scalars(candl_program_p program,
      / expanded / renamed. */
   for (i = 0; scalars[i] != -1; ++i)
     {
-      idx = scalars[i];
       /* Go to the first statement referencing the scalar. */
       for (j = 0; j < program->nb_statements; ++j)
 	if (candl_dependence_var_is_ref (program->statement[j], scalars[i])
@@ -2359,10 +2375,16 @@ candl_dependence_analyze_scalars(candl_program_p program,
 	      /* No more statement in the chain, exit. */
 	      if (stmts[0] == NULL)
 		break;
-
 	      int c = 0;
 	      is_priv = candl_dependence_var_is_ref (stmts[c], scalars[i])
 		== CANDL_VAR_IS_DEF;
+	      /* Ensure we have a use in the chain. */
+	      for (l = c + 1; stmts[l - 1] && stmts[l]; ++l)
+		if (candl_dependence_var_is_ref (stmts[l], scalars[i]) ==
+		    CANDL_VAR_IS_USED)
+		  break;
+	      if (stmts[l - 1] == NULL || stmts[l] == NULL)
+	      	is_priv = 0;
 	      /* Check for privatization, while the entry of the chain
 		 is a DEF. */
 	      while (stmts[c] && candl_dependence_var_is_ref
@@ -2429,16 +2451,16 @@ candl_dependence_analyze_scalars(candl_program_p program,
 			{
 			  free (program->scalars_privatizable);
 			  program->scalars_privatizable =
-			    (int*)malloc(priv_buff_size * 2 * sizeof(int));
+			    (int*)malloc(priv_buff_size * sizeof(int));
 			  for (l = 0; l < priv_buff_size; ++l)
 			    program->scalars_privatizable[l] = -1;
 			}
-		      if (2 * nb_priv == priv_buff_size)
+		      if (nb_priv == priv_buff_size)
 			{
 			  program->scalars_privatizable =
 			    realloc(program->scalars_privatizable,
 				    (priv_buff_size *= 2) * sizeof(int));
-			  for (l = 2 * nb_priv; l < priv_buff_size; ++l)
+			  for (l = nb_priv; l < priv_buff_size; ++l)
 			    program->scalars_privatizable[l] = -1;
 			}
 		      /* Memorize the scalar information in the
@@ -2733,7 +2755,7 @@ void candl_compute_last_writer (CandlDependence *dep, CandlProgram *prog)
 {
   // int count=0;
     while (dep != NULL)    {
-        if (dep->type == CANDL_RAW || dep->type == CANDL_WAW || dep->type == CANDL_RAR)   {
+        if (dep->type != CANDL_WAR)   {
 	  // printf("Last writer for dep %d: %d %d\n", count++, dep->source->depth, dep->target->depth);
             // candl_matrix_print(stdout, dep->domain);
             candl_dep_compute_lastwriter(&dep, prog);
