@@ -258,7 +258,7 @@ int *pluto_constraints_solve(PlutoInequalities *tmpcst, Stmt *stmts, int nstmts)
 
     /// LNP:
     constraints_free(newcst);
-    
+
     return fsol;
 
 }
@@ -629,7 +629,7 @@ int find_permutable_hyperplanes(PlutoProg *prog, int max_sols,
     }while (num_sols_found < max_sols && bestsol != NULL);
 
     free(orthcst);
-    
+
     /* same number of solutions are found for each stmt */
     return num_sols_found;
 }
@@ -1087,7 +1087,7 @@ void normalize_domains(PlutoProg *prog)
                 dpoly->val[dpoly->nrows-1][2*nvar+j] = 1;
             }
         }
-	
+
         /* LNP: update the reference, add_row can resize */
 	dep->dpolytope = dpoly;
     }
@@ -1196,7 +1196,7 @@ int pluto_auto_transform (PlutoProg *prog)
 	   * fusable loops */
 
 	  ddg_compute_scc(prog);
-	  
+
 	  if (ddg->num_sccs <= 1 || depth > 32)   {
 	    printf("Number of unsatisfied deps: %d\n", get_num_unsatisfied_deps(prog->deps, prog->ndeps));
 	    printf("Number of unsatisfied inter-stmt deps: %d\n",
@@ -1668,6 +1668,11 @@ Graph *ddg_create (PlutoProg *prog)
 		if (IS_RAR(dep->type)) continue;
 		/* remember it's a multi-graph */
 		g->adj->val[dep->src][dep->dest] += !dep_is_satisfied(dep);
+		/// LNP: Enforce non-distributability between statements that
+		/// correspond to a def-use chain for a scalar, to be
+		/// privatized.
+		if (dep->type == CANDL_RAW_SCALPRIV)
+		  g->adj->val[dep->dest][dep->src] += !dep_is_satisfied(dep);
 	}
 
 	return g;
